*  Program...........: LASSO.PRG v1.1
*  Author............: Richard D. Hodder (Rick)
*                      Compuserve 71201,1705
*
*  Description.......: GENSCRNX LASSO Driver
*  Draw a "Lasso" (box) around a set of screen objects and then put 
*  *:LASSO into the comment snippet of the "Lasso" and then put any 
*  GENSCRNX directives that you want to apply to the end of the comment 
*  snippet of each object whose UPPER LEFT CORNER falls within the 
*  dimensions of the "Lasso".  Options for: 
*   - excluding a specific object from being affected by ANY "Lasso"
*     (*:NOLASSO)
*   - defining which object TYPES (ex. pushbuttons) in a "lasso" will 
*     be affected (included/excluded) (*:LASSOINCLUDE, *:LASSOEXCLUDE)
*   - defining an expression which determines which objects within the 
*     lasso will be be affected (*:LASSOIF)
*   - DEBUG mode which leaves a "lasso" visible on the working screen 
*     after processing is complete (*:LASSODEBUG)
*   - DEBUG mode which leaves ALL "lasso"s visible on the working screen 
*     after processing is complete (*:LASSODEBUG)
*   - turning off driver in setup code (good during development) 
*     leaving all "lassos" visible in the working screen without 
*     processing "lassos" (*:NOLASSO)
*   - ability to define multiple lassos within the same lasso object
*   - ability to define "screen" lassos: lassos whose comment snippets
*     are applied to all screen objects (*:LASSO SCREEN).  Can also have
*     multiple screen lassos within the same lasso object.  Can also have
*     normal and screen lassos within the same lasso object.
*   - ability to define "screenset" lassos whose comment snippets are 
*     applied to the same area on all screens in the screenset.  You
*     can also define "screenset" screen lassos, that will affect all
*     objects on all screens in the screenset.
*   - ability to make the lasso object a screen object that remains
*     and can be affected by other overlapping lassos or other drivers
*     (ex. 3D) (*:LASSOSAVE)
*     ex. you want a box around a set of objects: rather than creating
*     a box, and then creating a lasso around the box and objects, just use 
*     a box for the lasso object and then put *:LASSOSAVE in the comment 
*     snippet.  The dimensions and LASSO directives of the lasso object 
*     will be used for the lasso, and the box will remain on the screen.
*     *:LASSOSAVE was an idea given to me by Steven Black (the creator
*     of the TABS driver, INTL, GENX.HLP file, as well as numerous articles
*     on GENSCRNX)
*   - ability to apply the lasso's comment snippet to the beginning
*     rather than the end of the working object comment snippet
*     on the CONFIG.FPx (_LASSOBEFORE="ON"),
*     Screen (*:LASSOBEFORE,*:LASSOAFTER), and Lasso (*:LASSO [BEFORE|AFTER]) 
*     level.
*   - ability to define a lasso or a group of lassos in an external file
*     and apply them to a screen (*:LASSOFILE <filename>.<ext>) OR to all 
*     screens in a project (CONFIG.FP - LASSOFILE="<filename>.<ext>"
*   - I have tried this driver in DOS and WINDOWS: I assume that it works
*     for Mac and UNIX: anyone using these 2 platforms, please give me 
*     feedback as to how well it works/problems encountered.
*
*  The lasso object does not appear in the SPR.  This driver will "lasso" 
*  any objects that are inserted with *:INSSCX, whose UPPER LEFT CORNER 
*  falls within the dimensions of the "Lasso".  Screens inserted with 
*  *:INSSCX can also contain "lassos" if the screen they are inserted into 
*  invokes the driver.
*
*  There can be many lassos, and they can overlap.
*  Lassos do NOT have to be boxes, they can be ANY screen object.    
*
* LASSO DIRECTIVES
*
* CONFIG.FP(x) DIRECTIVES
* -------------------------
*  LASSO="OFF"                     - Disables Lasso driver globally
*  LASSOBEFORE="ON"                - Global directive that prepends Lasso 
*                                    comment snippet to working object 
*                                    comment snippet.
*  LASSOFILE="<filename>.<ext>"    - ability to define a lasso or a group 
*                                    of lassos in an external file
*                                    and apply them to all screens in a project
*                                    LASSO assumes that all such lassos are
*                                    SCREEN lassos
*
* SETUP SNIPPET DIRECTIVES
* ------------------------
*  *:LASSODEBUG                    - Leaves all lasso objects on working 
*                                    screen after processing.  For debugging
*                                    purposes only: for permanent objects
*                                    use *:LASSOSAVE.
*  *:LASSOBEFORE                   - Prepend Lasso comment snippet to 
*                                    working object comment snippet for 
*                                    current screen only.  
*  *:LASSOAFTER                    - Appends Lasso comment snippet to 
*                                    end of working object comment snippet.
*                                    "AFTER" is the default, but this allows
*                                    for overide of CONFIG.FP *:LASBEFORE
*  *:LASSOFILE <filename>.<ext>    - ability to define a lasso or a group 
*                                    of lassos in an external file
*                                    and apply them to all screens in a project
*                                    LASSO assumes that all such lassos are
*                                    SCREEN lassos
*
*  *:NOLASSO                       - disable LASSO Driver
*    
* COMMENT SNIPPET DIRECTIVES
* --------------------------
*
*  *:LASSO [SCREEN] [BEFORE|AFTER] [SCREENSET] - Defines object as a Lasso
*                                   options for "screen" lassos, inserting
*                                   comment snippet BEFORE or AFTER the
*                                   existing comment snippet, and an option
*                                   for "SCREENSET" lassos
* *:LASSOINCLUDE <expN1>|<expC1> [,<expN2>|<expC2>..]
*                                  - List of object types to be affected by
*                                   Lasso.  Valid <expC>s are TEXT, BOX, 
*                                   LIST, PUSHBUTTON, RADIOBUTTON, CHECKBOX, 
*                                   SAY, GET, EDITREGION, POPUP, PICTURE, 
*                                   INVISIBLEBUTTON, SPINNER
*
* *:LASSOIF <expL>                - Logical expression determines screen 
*                                   objects to be affected by Lasso
* *:LASSOEXCLUDE <expN1>|<expC1> [,<expN2>|<expC2>..]
*                                 - List of object types that should NOT 
*                                   be affected by Lasso.   Valid <expC>s are TEXT, BOX, 
*                                   LIST, PUSHBUTTON, RADIOBUTTON, CHECKBOX, 
*                                   SAY, GET, EDITREGION, POPUP, PICTURE, 
*                                   INVISIBLEBUTTON, SPINNER
*    
* *:LASSODEBUG                    - Leaves lasso object on working screen
*                                   after processing.  NOTE: for debugging
*                                   ONLY: for permanent object use 
*                                   *:LASSOSAVE
* *:LASSOSAVE                     - Leaves the lasso object as a screen
*                                   object, rather than removing it.  This
*                                   allow the lasso object to be affected by 
*                                   other overlapping objects.
* *:NOLASSO                       - Prevents the screen object from being
*                                   affected ANY lassos
* *:LASSOADDBRACES <expC>         - frames whatever follows with {{ }}
*                                   ex. 
*                             *:LASSOADDBRACES &. REPLACE DISABLE WITH .T.
*                             gets converted to the following at generate
*                             time:
*                                {{&. REPLACE DISABLE WITH .T.}}
*
*    Please send any comments, ideas, improvements, suggestions, bug reports,
*    to Rick Hodder CIS: 71201,1705 
*                        FOXUSER forum section 9 - 3rd Party Products.
*
*    Please tell me how you are using this driver in your work, it will 
*    make future versions that much better.
*
*  With thanks to 
*     o Ken Levy [FLASH] for GENSCRNX,
*     o Steve Black [Steven Black Consulting / UP!] for GENX.HLP and 
*       numerous articles on GENSCRNX and suggestions for LASSO
*     o Randy Pearson [CYCLA] for the GENSCRNX Reference (GSXREF.ZIP), 
*       and contributions to GENX.HLP
*
*  Version...........: 1.1  -- February 18 1995
*  Copyright.........: (c) Richard D. Hodder, 1995
*          This software is provided "as is" without
*          any warranty expressed or implied. In no event shall
*          Richard D. Hodder, contributors, or distributors be liable
*          for any damages.
*
*          This software requires GENSCRNX 2.0, a public domain program 
*          created by Ken Levy at the JPL in Pasadena, California.
*
*          This software has been made PUBLIC DOMAIN by the author.
*
*
*  Dependencies......: FP 2.0 or greater: works on all platforms!
*                      GENSCRNX 2.0 or greater
*                      This driver MUST be run at SCXDRV1
*** CHANGE - changed all references of the __LASSO array to 
***          __LASSO&lcPlatform
***        - changed all references of the __LASDEBUG memory variable
***          __LASDB&lcPlatform
***        - changed all references of the __LASBEFORE memory variable
***          __LASB4&lcPlatform

PRIVATE lcDriver, c_LASSO, m_LASSO, c_LASEXCL, c_LASINCL, c_NOLASSO
PRIVATE c_LASDEBUG, c_LASBEFORE, c_LASAFTER, c_LASADDBRACES, c_LASSOIF
PRIVATE c_LASSOSAVE, c_LASSOFILE, lcINCLIST, lcEXCLIST
PRIVATE llAt1Setup, llAt1Config, llAt3Setup, llAt3Config
PRIVATE lnSCX1Cnt, lnSCX3Cnt, lnNumDRV1, lnNumDRV3, lcPlatform

M.lcDriver=PROGRAM()
m.lcPlatform=ALLTRIM(m.platform_)

* Check if driver was disabled in config.fp ( LASSO=OFF) or by 
* m._LASSO=OFF

IF NOT DrvEnable(PROGRAM())
  GOTO BOTTOM
  RETURN .F.
ENDIF

* Check for GENSCRNX 2.0 - I am not sure that 2.0 is the only version that 
* this driver works with, but I know the driver works in 2.0

IF NOT VAL(m.gsxversion) >= 2.0
  = Warning(PROGRAM() + " requires GENSCRNX v2.0.")
  GO BOTTOM
  RETURN .F.
ENDIF

* CHECK TO MAKE SURE THAT THE DRIVER IS BEING RUN AT SCXDRV1, 3, OR 8

IF !INLIST(m.drv_no,1,3,8)
  = Warning(PROGRAM() + " must be run at SCXDRV1.")
  GO BOTTOM
  RETURN .F.
ENDIF

* MAKE SURE THAT LASSO IS RUN ORIGINALLY AT SCXDRV1.  A CHECK HAS ALREADY 
* BEEN PERFORMED THAT MAKES SURE THAT LASSO IS NOT BEING RUN AT ANYTHING
* OTHER THAN 1, 3, OR 8.  

IF m.drv_no=3
   llAt1Config=.F.
   llAt1Setup=.F.
 
   lnSCX1Cnt=1
   DO WHILE configfp("_SCXDRV1","-",lnSCX1Cnt)!="-"
      IF "LASSO" $ UPPER(configfp("_SCXDRV1","-",lnSCX1Cnt))
         llAt1Config=.T.
         EXIT
      ELSE
         lnSCX1Cnt=lnSCX1Cnt+1 
      ENDIF
   ENDDO   

   lnNumDRV1=WSOCCURS("*:SCXDRV1",.T.)
   IF lnNumDRV1!=0
      FOR I=1 TO lnNumDRV1
          IF "LASSO" $ UPPER(WORDSEARCH("*:SCXDRV1",.T.,.F.,.F.,I))
             llAt1Setup=.T.
          ENDIF
      NEXT I   
   ENDIF   

   IF !( llAt1Config .OR. llAt1Setup)
      =WARNING("LASSO must run at SCXDRV1 either in Setup OR Config.FPx.")  
      =CLEANUP()
      CANCEL
   ENDIF
ENDIF

m.c_LASSO='*:LASSO'
m.m_LASSO='*-:LASSO'
m.c_LASINCL='*:LASSOINCLUDE'
m.c_LASEXCL='*:LASSOEXCLUDE'
m.c_NOLASSO='*:NOLASSO'
m.c_LASDEBUG='*:LASSODEBUG'
m.c_LASBEFORE='*:LASSOBEFORE'
m.c_LASAFTER='*:LASSOAFTER'     && OVERRIDE OF LASBEFORE
m.c_LASADDBRACES='*:LASSOADDBRACES'
m.c_LASSOIF='*:LASSOIF'
m.c_LASSOSAVE='*:LASSOSAVE'

*** CHANGE: ADD DIRECTIVES FOR *:LASSOFILE

m.c_LASSOFILE='*:LASSOFILE'

*** ENDCHANGE

* IF "*:NOLASSO" IS IN SETUP CODE, THEN DO NOT EXECUTE DRIVER

IF WORDSEARCH(m.c_NOLASSO,.T.)!=CHR(0)
   GO BOTTOM 
   RETURN
ENDIF

* IDEA: DRIVER GETS INVOKED 3 LEVELS  
*       1. FIRST AT SCXDRV1 TO CREATE AN ARRAY (__LASSO) TO STORE 
*          HPOS, VPOS, WIDTH, HEIGHT, AND COMMENT SNIPPET, AND 
*          INCLUDE/EXCLUDE DIRECTIVES, AND THEN PUT "*:SCXDRV3 LASSO"
*          AND "*:SCXDRV8 LASSO" IN THE SETUP SNIPPET OF THE SCREEN
*
*       2. NEXT AT SCXDRV3 TO APPLY THE COMMENT SNIPPET TO THE END 
*          OF THE COMMENT SNIPPET OF ANY OBJECT WHOSE UPPER LEFT CORNER
*          FALLS INSIDE THE LASSOS DEFINED IN __LASSO.
*
*       3. NEXT AT SCXDRV8 TO RELEASE __LASSO ARRAY

*on key label escape
*set escape on
   
* RUN THE ROUTINE BASED ON WHAT THE SCXDRV LEVEL IS

DO CASE
   CASE m.drv_no=1
        DO LEVEL1        
   CASE m.drv_no=3
        DO LEVEL3        
   CASE m.drv_no=8
        DO LEVEL8
ENDCASE

GO BOTTOM
RETURN

FUNCTION BldList
PARAMETERS pcSNIPPET

* PURPOSE:
*    THIS FUNCTION PARSES pcSNIPPET, FINDING AND REMOVING *:LASSOINCLUDE, 
*    *:LASSOIF, AND *:LASSOEXCLUDE, CONVERTING THEM INTO A SINGLE PROGRAM 
*    STATEMENT WHICH IS RETURNED AT THE END OF ROUTINE.
*
*    NOTE: IF BOTH *:LASSOINCLUDE AND *:LASSOEXCLUDE ARE IN pcSNIPPET,
*          ONLY THE INCLUDE WILL BE PROCESSED.
*
* EX.  pcSNIPPET CONTAINS        *:LASSO
*                                *:LASSOINCLUDE 14,15
*                                *:LASSOIF X>5
*    
*      RETURNS " .AND. ( INLIST(OBJTYPE,14,15) ) .AND. ( X>5 ) "

PRIVATE lcRetVal,lcExpr,lcRemLine,dummy
lcRetVal=""
lcRemLine=""
dummy=0

IF WORDSEARCH(m.c_LASINCL,pcSNIPPET)=CHR(0) .AND. ;
   WORDSEARCH(m.c_LASEXCL,pcSNIPPET)=CHR(0)  .AND. ;
   WORDSEARCH(m.c_LASSOIF,pcSNIPPET)=CHR(0)

   RETURN " "   
ENDIF

* GRAB EXPRESSION FOLLOWING "*:LASSOIF" 

lcExpr=""
IF WORDSEARCH(m.c_LASSOIF,pcSNIPPET)!=CHR(0) 
   &pcSNIPPET=RemLine(m.c_LASSOIF,pcSNIPPET,@lcRemLine)   
   lcExpr=STRTRANC(lcRemLine,c_LASSOIF,"")
ENDIF

* ASSUME ALL OBJECTS INCLUDED AT FIRST, THEN APPLY EXCLUSIONS AND INCLUSIONS

IF WORDSEARCH(m.c_LASEXCL,pcSNIPPET)!=CHR(0) 
   lcRetVal=""
   lcEXCLIST=WORDSEARCH(m.c_LASEXCL,pcSNIPPET)   
   &pcSNIPPET=RemLine(m.c_LASEXCL,pcSNIPPET)
   lcRetVal=ALLTRIM(BldSpecial(@lcEXCLIST))

   IF EMPTY(lcRetVal)
      IF EMPTY(lcEXCLIST)
         lcRetVal=""
      ELSE
         lcRetVal=" .AND. !INLIST(OBJTYPE,"+lcEXCLIST+")"
      ENDIF
   ELSE
      IF EMPTY(lcEXCLIST)
         lcRetVal=" .AND. !("+lcRetVal+")"
      ELSE
         lcRetVal=" .AND. !("+lcRetVal+" .OR. INLIST(OBJTYPE,"+lcEXCLIST+") )"
      ENDIF
   ENDIF
 
ENDIF

IF WORDSEARCH(m.c_LASINCL,pcSNIPPET)!=CHR(0) 
   lcRetVal=""
   lcINCLIST=WORDSEARCH(m.c_LASINCL,pcSNIPPET)
   &pcSNIPPET=RemLine(m.c_LASINCL,pcSNIPPET)
   
   lcRetVal=ALLTRIM(BldSpecial(@lcINCLIST))

   IF EMPTY(lcRetVal)
      IF EMPTY(lcINCLIST)
         lcRetVal=""
      ELSE
         lcRetVal=" .AND. INLIST(OBJTYPE,"+lcINCLIST+") "
      ENDIF
   ELSE
      IF EMPTY(lcINCLIST)
         lcRetVal=" .AND. ("+lcRetVal+")"
      ELSE
         lcRetVal=" .AND. ("+lcRetVal+" .OR. INLIST(OBJTYPE,"+lcINCLIST+") )"
      ENDIF
   ENDIF
   
ENDIF

IF !EMPTY(lcExpr)
   lcRetVal=lcRetVal+" .AND. ("+lcExpr+")"
ENDIF
   

RETURN lcRetVal

FUNCTION BldSpecial
PARAMETER pcList
* PARAMETERS: pcList - A CHARACTER VARIABLE PASSED BY REFERENCE (USING @)
*                      CONTAINING A LIST OF NUMBERS SEPARATED BY COMMAS
*
* PURPOSE: 
*  THIS FUNCTION PARSES pcList AS FOLLOWS:
*     1  REMOVES SPACES
*     2. DOES A SPECIAL CONVERSION (INSPIRED BY STEVEN BLACK'S REFERRAL TO 
*        OBJTYPE 15 SCX RECORDS AS 15.0, 15.1, AND 15.2 IN GENX.HLP) OF ANY
*        NUMBERS WITH DECIMAL POINTS IN THEM.  THE NUMBER TO THE LEFT OF THE
*        DECIMAL POINT IS CONSIDERED THE OBJECT TYPE (OBJTYPE) AND THE
*        NUMBER TO THE RIGHT OF THE DECIMAL POINT IS CONSIDERED THE 
*        OBJECT CODE (OBJCODE).  THESE DECIMAL NUMBERS ARE REMOVED FROM
*        pcList AND THEN ARE CONVERTED TO A LOGICAL EXPRESSION STORED INSIDE 
*        THE VARIABLE lcSpecial SEPARATED WITH ".OR."s.
*
*  EX. IF pcList IS  "13,14,15.0,15.2,16", THEN AT THE END OF THIS FUNCTION,
*      pcList CONTAINS "13,14,16" AND lcSpecial CONTAINS
*      " (OBJTYPE=15 .AND. OBJCODE=0) .OR. (OBJTYPE=15 .AND. OBJCODE=2) "
*
*      lcSpecial IS THE VALUE RETURNED BY THIS FUNCTION.  IF THERE IS 
*      NO DECIMAL POINTS IN pcList, THE FUNCTION RETURNS ""
*
* NOTE!: NO ERROR CHECKING IS PERFORMED ON THESE NUMBERS, IN ORDER TO
*        ALLOW FOR ANY FUTURE OBJECT TYPES THAT HAVE MULTIPLE OBJECT CODES 
* 
* llFirst:  ON FIRST PASS, THERE SHOULD NOT BE AN ".OR." AT THE BEGINNING
*           OF lcSpecial: llFirst IS USED TO PREVENT THIS FROM HAPPENING
* lnPrdLoc: LOCATION OF PERIOD IN pcList
* lnStart:  LOCATION OF FIRST COMMA TO THE LEFT OF lnPrdLoc
* lnStop:   LOCATION OF FIRST COMMA TO THE RIGHT OF lnPrdLoc

PRIVATE llFirst,lnPrdLoc,lnStart,lcSpecial

llFirst=.T.
lcSpecial=""

* REMOVE SPACES FROM LIST OF VALUES, CONVERT OBJECT LIST TO UPPER CASE,
* AND CONVERT ALL WORDS IN THE OBJECT LIST TO CORRESPONDING NUMBERS

pcList=UPPER(STRTRAN(pcList," "))

pcList=CnvrtWrd(pcList)

* ADD COMMA TO END OF LIST

IF RIGHT(pcList,1)!=","
   pcList=pcList-","
ENDIF   

DO WHILE "." $ pcList
   lnPrdLoc=AT(".",pcList)
   IF lnPrdLoc=1
        = Warning(PROGRAM() + ": LASINCL/LASEXCL error: "+m.cr_lf+ ;
                  "Use of '.' incorrect.  Items in list must be of the form:"+m.cr_lf+ ;
                  "OBJTYPE.OBJCODE     ex 15.2")
        =cleanup()
        cancel          
   ENDIF
   
   lnStart=lnPrdLoc-1
   DO WHILE lnStart>1 .AND. !(SUBSTR(pcList,lnStart,1) $ ",") 
      lnStart=lnStart-1
   ENDDO   
   lnStart=IIF(lnStart=1,0,lnStart)

   lnStop=lnPrdLoc+1
   DO WHILE lnStop<LEN(pcList) .AND. !(SUBSTR(pcList,lnStop,1) $ ",") 
      lnStop=lnStop+1
   ENDDO

   lcSpecial=lcSpecial+IIF(llFirst," "," .OR. ")+ ;
             "(OBJTYPE="+SUBSTR(pcList,lnStart+1,lnPrdLoc-lnStart-1)+ ;
             " .AND. OBJCODE="+SUBSTR(pcList,lnPrdLoc+1,lnStop-lnPrdLoc-1)+") "
             
   * llFirst SET TO .F. SO THAT ANY FUTURE PASSES WILL HAVE ".OR." PREPENDED

   llFirst=.F.          

   * REMOVE DECIMAL NUMBER FROM pcList
   
   pcList=IIF(lnStart!=1,SUBSTR(pcList,1,lnStart),"")+ ;
          SUBSTR(pcList,lnStop+1)          
   
ENDDO
   
* REMOVE RIGHTMOST COMMA FROM pcList

DO WHILE RIGHT(pcList,1)=","
   pcList=SUBSTR(pcList,1,LEN(pcList)-1)
ENDDO

RETURN lcSpecial

PROCEDURE LEVEL1

* IN CASE THE DRIVER FAILED PREVIOUSLY AND LEFT THE __LASSO ARRAY,
* __LASDEBUG, AND __LASBEFORE IN MEMORY, REMOVE THEM.  BUT SINCE SCREENSET
* LASSOS REMAIN IN THE __LASSO ARRAY, IT SHOULD ONLY BE RELEASED WHEN 
* PROCESSING THE 1ST SCREEN OF A SCREENSET

IF m.scxcount=1
***c      
   IF TYPE("__LASSO"+lcPlatform)!="U"
      RELEASE __LASSO&lcPlatform
   ENDIF   

ENDIF
***c
IF TYPE("__LASDB"+lcPlatform)!="U"
   RELEASE __LASDB&lcPlatform
ENDIF   

IF TYPE("__LASB4"+lcPlatform)!="U"
   RELEASE __LASB4&lcPlatform
ENDIF   
   
* MOVE TO SCREEN HEADER RECORD TO CHECK FOR:
*   IS THIS A LIBRARY SCREEN (DO NOT PROCESS)
*   IS THE DRIVER TURNED OFF IN THE SETUP WITH *:NOLASSO 
*   CHECK IF DEBUG MODE FOR DRIVER IS ON (LEAVE LASSO BOXES ON SCREEN)
*   CHECK IF BEFORE MODE (COMMENT OF LASSO INSERTED BEFORE RATHER THAN
*      AFTER THE COMMENT OF THE WORKING SCREEN OBJECT) IS ON
*     STUFF CALLS TO SCXDRV3 AND SCXDRV8 INTO SETUPCODE

LOCATE FOR OBJTYPE=1 .AND. !DELETED()
 
* IF THIS IS A LIBRARY SCREEN, OR A TEMPLATE SCREEN DO NOT EXECUTE DRIVER
  
IF WORDSEARCH("*:DEFLIB",.T.)!=CHR(0) .OR. ;
   WORDSEARCH("*:NOXGEN",.T.)!=CHR(0) .OR. ;
   WORDSEARCH("*:NOGEN",.T.)!=CHR(0) 
      
   GO BOTTOM
   RETURN
ENDIF   
   
* IF *:NOLASSO IS IN SETUP THEN DO THE FOLLOWING TO ALL LASSOS IN SCREEN
     * REPLACE COMMENT WITH "*:NO3D"
     * BLANK OUT VALID, WHEN, RANGELO, AND RANGEHI OF LASSOS
     * SO THAT LASSOS BECOME VISIBLE, NON-3D, HARMLESS OBJECTS WITHOUT 
     * GENSCRNX DIRECTIVES (EXCEPT FOR *:NO3D)
     * DISABLE LASSO BY REPLACING DISABLE WITH .T.
   * AND DO NOT EXECUTE DRIVER ANY FURTHER

IF WORDSEARCH(m.c_NOLASSO,.T.)!=CHR(0) 
   SCAN ALL FOR WORDSEARCH(m.c_LASSO)!=CHR(0) AND !DELETED()
        REPLACE COMMENT WITH "*:NO3D"+m.cr_lf, ;
                VALID WITH "", ;
                WHEN WITH "", ;
                RANGELO WITH "", ;
                RANGEHI WITH "", ;
                DISABLED WITH .T.
   ENDSCAN
   GO BOTTOM
   RETURN
ENDIF     

* IF "*:LASSODEBUG" IS IN SETUPCODE,   
* THEN SET GLOBAL FLAG THAT WILL LEAVE LASSO BOX SO THAT YOU CAN SEE
* WHAT YOU HAVE DONE
   
PUBLIC __LASDB&lcPlatform
STORE .F. TO __LASDB&lcPlatform
IF WORDSEARCH(m.c_LASDEBUG,.T.)!=CHR(0)
   __LASDB&lcPlatform=.T.    
ENDIF
   
PUBLIC __LASB4&lcPlatform
STORE .F. TO __LASB4&lcPlatform
   
* CHECK IF "BEFORE" MODE IS IN CONFIG.FP ( LASSOBEFORE="ON" )

IF drvenable("LASSOBEFORE")
   __LASB4&lcPlatform=.T.   
ENDIF
   
* CHECK IF "BEFORE" MODE IS IN SETUPCODE 

IF WORDSEARCH(m.c_LASBEFORE,.T.)!=CHR(0)
   __LASB4&lcPlatform=.T.   
ENDIF
   
* CHECK IF "AFTER" MODE IS IN SETUPCODE (THIS IS TO ALLOW AN OVERIDE
* TO SETTING LASSOBEFORE="ON")

IF WORDSEARCH(m.c_LASAFTER,.T.)!=CHR(0)
   __LASB4&lcPlatform=.F.   
ENDIF

*** CHANGE: ADDED CODE FOR THE LASSOFILE CONFIG.FP DIRECTIVE AND THE 
***         *:LASSOFILE SETUP DIRECTIVE.  CHECK FOR EXISTENCE OF 
***         THE FILE, AND IF FOUND ADD LASSO TO THE __LASSO ARRAY

IF WORDSEARCH(m.c_LASSOFILE,.T.)!=CHR(0) .OR. ; 
   configfp("LASSOFILE","-",1)!="-"

   lcOldAlias=ALLTRIM(STR(SELECT()))
   CREATE CURSOR LASSOTMP (COMMENT M)
   APPEND BLANK

   IF configfp("LASSOFILE","-",1)!="-"
      lcLASSOFILE=configfp("LASSOFILE","-",1)
      IF !FILE(lcLASSOFILE)
         =WARNING("CONFIG.FP LASSO FILE '"+lcLASSOFILE+"' could not be found")  
         SELECT &lcOldAlias
         =CLEANUP()
         CANCEL
      ENDIF
      APPEND MEMO COMMENT FROM (lcLASSOFILE) OVERWRITE
      DO PROCLASSO WITH "CONFIG"
   ENDIF
   
   IF WORDSEARCH(m.c_LASSOFILE,.T.)!=CHR(0)
      lcLASSOFILE=WORDSEARCH(m.c_LASSOFILE,.T.)
      IF !FILE(lcLASSOFILE)
         =WARNING("SETUP LASSO FILE '"+lcLASSOFILE+"' could not be found")  
         SELECT &lcOldAlias
         =CLEANUP()
         CANCEL
      ENDIF
      APPEND MEMO COMMENT FROM (lcLASSOFILE) OVERWRITE
      DO PROCLASSO WITH "SETUP"
   ENDIF

   * CLOSE CURSOR AND RESTORE WORKAREA

   USE
   SELECT &lcOldAlias
   
ENDIF

*** ENDCHANGE

* THE LASSO DRIVER WORKS AT SCXDRV LEVELS 1, 3 AND 8.  IF YOU DO NOT SPECIFY
* AN SCXDRV3 CALL TO LASSO, THEN "*:SCXDRV3 LASSO" IS STUFFED AT THE TOP OF 
* THE SETUP CODE, THIS MAKES LASSO THE FIRST SCXDRV3 DRIVER TO EXECUTE.  
* IF YOU FIND THAT YOUR SCXDRV3 DRIVER NEEDS TO EXECUTE BEFORE LASSO, 
* THEN PUT:
*     *:SCXDRV3 YOURDRIVER
*     *:SCXDRV3 LASSO
* INTO THE SETUP CODE OR
*     _SCXDRV3 YOURDRIVER
*     _SCXDRV3 LASSO
* INTO YOUR CONFIG.FPx
*
* THIS DRIVER AUTOMATICALLY STUFFS "*:SCXDRV8 LASSO" INTO THE SETUP CODE.
* WHEN LASSO IS CALLED AT SCXDRV8, THE VARIABLES (SUCH AS THE __LASSO ARRAY)
* ARE RELEASED.

llAt3Config=.F.
llAt3Setup=.F.
 
lnSCX3Cnt=1
DO WHILE configfp("_SCXDRV3","-",lnSCX3Cnt)!="-"
   IF "LASSO" $ UPPER(configfp("_SCXDRV3","-",lnSCX3Cnt))
      llAt3Config=.T.
      EXIT
   ELSE
      lnSCX3Cnt=lnSCX3Cnt+1 
   ENDIF
ENDDO   

lnNumDRV3=WSOCCURS("*:SCXDRV3",.T.)
IF lnNumDrv3!=0
   FOR I=1 TO lnNumDRV3
       IF "LASSO" $ UPPER(WORDSEARCH("*:SCXDRV3",.T.,.F.,.F.,I))
         llAt3Setup=.T.
      ENDIF
   NEXT I   
ENDIF   

IF ( llAt3Setup .OR. llAt3Config )
   * THERE IS ALREADY A CALL AT SCXDRV3, SO JUST 
   * STUFF A CALL TO THE LASSO DRIVER AT SCXDRV8 AT THE TOP OF
   * THE SETUPCODE 

   =WORDSTUFF(M.cr_lf+"*:SCXDRV8"+lcDriver+m.cr_lf, ;
              .T.,.T.,.T.)
   
ELSE
   * STUFF A CALL TO THE LASSO DRIVER AT SCXDRV3 AND SCXDRV8 AT THE TOP OF
   * THE SETUPCODE (MAKING THE DRIVER THE FIRST ONE CALLED AT EACH LEVEL)

   =WORDSTUFF(M.cr_lf+"*:SCXDRV3"+m.lcDriver+ ;
              M.cr_lf+"*:SCXDRV8"+lcDriver+m.cr_lf, ;
              .T.,.T.,.T.)
ENDIF

             
* CHECK FOR ANY LASSOs DEFINED ON THE SCREEN (*:LASSO IN COMMENT SNIPPET),
* INITIALIZE/UPDATE __LASSO - AN ARRAY THAT KEEPS TRACK OF THE DIMENSIONS 
* AND COMMENT SNIPPET OF ALL LASSOS

LOCATE FOR WORDSEARCH(m.c_LASSO)!=CHR(0)

IF !FOUND()
   RETURN
ENDIF   

DO LEVEL3

PROCEDURE LEVEL3
PRIVATE I,lcINLIST

* CHECK FOR ANY LASSOS THAT WERE IN NEWLY INSERTED SCREENS

* CHECK FOR ANY LASSOs DEFINED ON THE SCREEN (*:LASSO IN COMMENT SNIPPET),
* INITIALIZE/UPDATE __LASSO - AN ARRAY THAT KEEPS TRACK OF THE DIMENSIONS 
* AND COMMENT SNIPPET OF ALL LASSOS

SCAN FOR WORDSEARCH(m.c_LASSO)!=CHR(0)

     DO PROCLASSO WITH "NORMAL"

ENDSCAN

* IF NO LASSOS DEFINED, NO NEED TO PROCESS THEM
***c
IF TYPE("__LASSO"+lcPlatform)="U"
   RETURN
ENDIF

* APPLY ALL LASSOS TO SCREEN OBJECTS, AVOIDING "NON-SCREEN-OBJECT" OBJECT
* TYPES 1,2,3,4,6,8,9,10,18,19,21,23  - YOU COULDNT PUT A BOX AROUND 
* THEM IF YOU WANTED TO!

FOR I=1 TO ALEN(__LASSO&lcPlatform,1)
    lcINLIST=__LASSO&lcPlatform[I,6]
  
    * IF THE LASSO IS A "SCREEN" LASSO (A LASSO THAT WILL BE 
    * APPLIED TO EVERY OBJECT ON THE SCREEN EVEN IF IT DOES NOT FALL 
    * WITHIN THE BOUNDS OF A LASSO) NO NEED TO CHECK HPOS AND VPOS.  
    * SINCE "SCREEN" AND "SCREENSET" ARE SIMILAR, TESTING FOR "SCREEN" 
    * MUST REMOVE "SCREENSET" FROM __LASSO[*,8] BEFORE TESTING FOR THE 
    * EXISTENCE OF "SCREEN"
    *
    * OTHERWISE CHECK HPOS & VPOS OF ALL OBJECTS TO SEE IF UPPER LEFT 
    * CORNER FALLS WITHIN THE BOUNDARIES OF LASSO 

    IF "SCREEN" $ STRTRAN(UPPER(__LASSO&lcPlatform[I,8]),"SCREENSET","")
       SCAN ALL FOR ;
                !DELETED() .AND. ;
                !INLIST(OBJTYPE,1,2,3,4,6,8,9,10,18,19,21,23) .AND. ;
                WORDSEARCH(m.c_LASSO)=CHR(0) .AND. ;
                WORDSEARCH(M.c_NOLASSO)=CHR(0) .AND. ;
                WORDSEARCH(M.m_LASSO+ALLTRIM(STR(I)))=CHR(0) .AND. ;
                WORDSEARCH(M.c_insobj)=CHR(0) .AND. ;
                WORDSEARCH(M.c_instxt)=CHR(0) .AND. ;
                WORDSEARCH(M.c_insscx)=CHR(0) ;
                &lcINLIST

            =ApplyLasso(I)
         
            * A CHANGE WAS MADE TO A SCREEN OBJECT, SO FORCE THE 
            * REPROCESSING LOOP TO REPEAT - SEE Randy Pearson's 
            * GENSCRNX Reference or Steven Black's GENX.HLP 2.6: 
            * topics Detailed Analysis / Hooks for more information
       
            m.loop_obj=.T.
        
       ENDSCAN
    ELSE
       SCAN ALL FOR ;
                !DELETED() .AND. ;
                !INLIST(OBJTYPE,1,2,3,4,6,8,9,10,18,19,21,23) .AND. ;
                BETWEEN(HPOS,__LASSO&lcPlatform[I,1],__LASSO&lcPlatform[I,1]+__LASSO&lcPlatform[I,3]) .AND. ;
                BETWEEN(VPOS,__LASSO&lcPlatform[I,2],__LASSO&lcPlatform[I,2]+__LASSO&lcPlatform[I,4]) .AND. ;
                WORDSEARCH(m.c_LASSO)=CHR(0) .AND. ;
                WORDSEARCH(M.c_NOLASSO)=CHR(0) .AND. ;
                WORDSEARCH(M.m_LASSO+ALLTRIM(STR(I)))=CHR(0) .AND. ;
                WORDSEARCH(M.c_insobj)=CHR(0) .AND. ;
                WORDSEARCH(M.c_instxt)=CHR(0) .AND. ;
                WORDSEARCH(M.c_insscx)=CHR(0) ;
                &lcINLIST

            =ApplyLasso(I)
         
            * A CHANGE WAS MADE TO A SCREEN OBJECT, SO FORCE THE REPROCESSING
            * LOOP TO REPEAT - SEE Randy Pearson's GENSCRNX Reference or
            * Steven Black's GENX.HLP 2.6: topics Detailed Analysis / Hooks
            * for more information
         
            m.loop_obj=.T.
        
       ENDSCAN
    ENDIF   
NEXT I

PROCEDURE ApplyLasso
PARAMETER pnLasso

* PURPOSE: THIS PROCEDURE CONTAINS CODE WHICH GETS EXECUTED FOR EACH
*          LASSO, REGARDLESS OF MOST PARAMETERS

* IF "BEFORE MODE" IS ON, PREPEND LASSO COMMENT SNIPPET TO WORKING
* OBJECT COMMENT SNIPPET, OTHERWISE ADD IT TO THE END OF THE 
* WORKING OBJECT COMMENT SNIPPET.  __LASSO[pnLasso,7]=.T. MEANS "BEFORE MODE"

IF __LASSO&lcPlatform[pnLasso,7]
   REPLACE COMMENT WITH EVLTXT(__LASSO&lcPlatform[pnLasso,5])+m.cr_lf+COMMENT
ELSE
   REPLACE COMMENT WITH COMMENT+M.cr_lf+EVLTXT(__LASSO&lcPlatform[pnLasso,5]+m.cr_lf)
ENDIF           

* STUFF "*-:LASSO" FOLLOWED BY THE ARRAY ROW OF THE LASSO IN 
* __LASSO INTO THE BOTTOM OF THE COMMENT SNIPPET SO THAT 
* OBJECTS DONT GET EVALUATED BY THE SAME LASSO MORE THAN ONCE, BUT
* ARE STILL LEFT AVAILABLE TO BE PROCESSED AS PART OF ANOTHER 
* LASSO.
*    EX.  *-:LASSO1 IS PUT INTO THE COMMENT SNIPPET OF AN OBJECT
*         JUST PROCESSED BY __LASSO[1] 
   
=WORDSTUFF(M.cr_lf+m.m_LASSO+ALLTRIM(STR(pnLasso))+M.cr_lf,.T.,.F.)

RETURN

PROCEDURE LEVEL8
PRIVATE lnLasso, lnNumLassos

* PROCESSING IS FINISHED FOR THE SCREEN AS FAR AS LASSO IS CONCERNED.  
* REMOVE ANY NON-SCREENSET LASSOS FROM __LASSO, LEAVING SCREENSET LASSOS
* IN __LASSO TO BE PASSED ON TO NEXT SCREEN IN THE SCREENSET

IF TYPE("__LASSO"+lcPlatform)="U"
   lnNumLassos=0
ELSE   
   lnNumLassos=ALEN(__LASSO&lcPlatform,1)
   FOR lnLasso=ALEN(__LASSO&lcPlatform,1) TO 1 STEP -1
       IF !("SCREENSET" $ __LASSO&lcPlatform[lnLasso,8])
          =ADEL(__LASSO&lcPlatform,lnLasso)
          lnNumLassos=lnNumLassos-1
       ENDIF
   NEXT lnLasso
ENDIF

IF lnNumLassos!=0
   DIME __LASSO&lcPlatform[lnNumLassos,8]
ENDIF

* IF THERE ARE NO SCREENSET LASSOS DEFINED, OR THIS IS THE LAST SCREEN IN 
* IN A SCREENSET, RELEASE THE VARIABLES

IF lnNumLassos=0 .OR. m.scxcount=m.screens 

   * RELEASE THE __LASSO ARRAY

   IF TYPE("__LASSO"+lcPlatform)!="U"
      RELEASE __LASSO&lcPlatform
   ENDIF

   * RELEASE THE DEBUG FLAG

   IF TYPE("__LASDB"+lcPlatform)!="U"
      RELEASE __LASDB&lcPlatform
   ENDIF

   * RELEASE THE BEFORE FLAG

   IF TYPE("__LASB4"+lcPlatform)!="U"
      RELEASE __LASB4&lcPlatform
   ENDIF
ENDIF
RETURN

FUNCTION WSOCCURS
PARAMETERS pcSearchStr, pcVar
* THIS FUNCTION IS A WORDSEARCH VERSION OF THE FOXPRO "OCCURS()"  FUNCTION.
* IT RETURNS THE NUMBER OF LINES BEGINNING WITH THE SEARCH STRING
* THE TWO PARAMETERS ARE THE SAME AS WOULD BE USED FOR WORDSEARCH

PRIVATE lnCnt
lnCnt=0

DO WHILE WORDSEARCH(pcSearchStr,pcVar,.F.,.F.,lnCnt+1)!=CHR(0)
   lnCnt=lnCnt+1
ENDDO

RETURN lnCnt


PROCEDURE PROCLASSO
* PURPOSE:
*    THIS PROCEDURE PROCESSES EACH LASSO OBJECT AS FOLLOWS:
*    1. CUTS OUT EVERYTHING FROM THE FIRST *:LASSO FOUND UNTIL THE
*       END OF THE COMMENT SNIPPET AND PUTS IT INTO lcCutOut
*    2. COUNT HOW MANY LASSOs ARE DEFINED IN lcCutOut
*    3. FOR EACH LASSO IN lcCutOut:
*       4. CUT OUT FROM BEGINNING OF LASSO UP UNTIL NEXT *:LASSO AND 
*          PUT INTO lcThisLasso
*       5. ADD A ROW TO THE __LASSO ARRAY AND STORE THE HPOS,VPOS, 
*          HEIGHT, WIDTH OF THE LASSO IN lcThisLasso INTO THE ARRAY.  
*          CONVERT INCLUDES/EXCLUDES AND LASSOIFs INTO A PROGRAM STATEMENT,           

*** CHANGE: ADDED A PARAMETER pcLASSOFILE TO DESIGNATE IF THE LASSO BEING 
***         PROCESSED IS FROM A FILE: "CONFIG" IF THE LASSO FILE IS DEFINED
***         IN THE CONFIG.FP. "SETUP" IF THE LASSO FILE IS DEFINED IN THE
***         SETUP SNIPPET. "NORMAL" FOR LASSOS DEFINED ON THE SCREEN.

PARAMETER pcLASSOFILE

*** END CHANGE

PRIVATE lnLassoPos, dummy, lcCutOut, lcThisLasso, lnNextLas, lnLassoCnt
PRIVATE lnNumLassos, llSaveLasso, llDebugLasso
M.DUMMY=0

*** CHANGE: CHECK FOR *:INSOBJs IN lasso: THEY CAUSE LASSO TO GO INTO AN 
***         INFINITE LOOP:

IF WORDSEARCH(c_INSOBJ)!=CHR(0)
   =WARNING("*:INSOBJ cannot be used in a lasso.")  
   =CLEANUP()
   CANCEL    
ENDIF

*** ENDCHANGE

llSaveLasso=.F.
llDebugLasso=.F.

*** CHANGE: LASSOS FROM A FILE HAVE NO SCREEN COORDINATES AND THEREFORE 
***         CANNOT BE "SAVED" OR "DEBUGGED", SO IGNORE *:LASSOSAVE & *:LASSODEBUG

IF pcLASSOFILE="NORMAL"
   IF WORDSEARCH(c_LASSOSAVE)!=CHR(0)
      REPLACE COMMENT WITH RemLine(c_LASSOSAVE,"COMMENT")   
      llSaveLasso=.T.
   ENDIF   

   IF WORDSEARCH(c_LASDEBUG)!=CHR(0)
      REPLACE COMMENT WITH RemLine(c_LASDEBUG,"COMMENT")
      llDebugLasso=.T.
   ENDIF   
ENDIF

*** ENDCHANGE

lnLassoPos=WORDSEARCH(c_LASSO,.F.,.F.,@m.dummy,1)
lcCutOut=SUBSTR(COMMENT,lnLassoPos)
REPLACE COMMENT WITH SUBSTR(COMMENT,1,lnLassoPos-1)
lnLassoPos=1
lnNumLassos=WSOCCURS(c_LASSO,"m.lcCutOut")
FOR I=1 to lnNumLassos
    IF lnNumLassos=1 OR I=lnNumLassos
       lcThisLasso=lcCutOut
       lcCutOut=""
    ELSE
       lcTempCut=SUBSTR(lcCutOut,ATC(m.c_LASSO,lcCutOut)+LEN(m.c_LASSO))
       llDone=.F.
       lnCnt=1
       lnPos=0
       DO WHILE .NOT. llDone
          lnPos=ATC(m.c_LASSO,lcTempCut,lnCnt)
          IF SUBSTR(lcTempCut,lnPos+LEN(m.c_LASSO),1)$m.CR_LF+" "
             llDone=.T.
          ELSE 
             lnCnt=lnCnt+1   
          ENDIF
       ENDDO
       lnNextLas=lnPos+LEN(m.c_LASSO)
       lcThisLasso=SUBSTR(lcCutOut,1,lnNextLas-1)
       lcCutOut=SUBSTR(lcCutOut,lnNextLas)
    ENDIF   
    IF TYPE("__LASSO"+lcPlatform)="U"
       PUBLIC __LASSO&lcPlatform
       DIMENSION __LASSO&lcPlatform[1,8]
    ELSE      
       DIMENSION __LASSO&lcPlatform[ALEN(__LASSO&lcPlatform,1)+1,8]
    ENDIF   
    lnLassoCnt=ALEN(__LASSO&lcPlatform,1)
    __LASSO&lcPlatform[lnLassoCnt,1]=HPOS
    __LASSO&lcPlatform[lnLassoCnt,2]=VPOS
    __LASSO&lcPlatform[lnLassoCnt,3]=WIDTH
    __LASSO&lcPlatform[lnLassoCnt,4]=HEIGHT
    __LASSO&lcPlatform[lnLassoCnt,6]=BldList("m.lcThisLasso")


    * COLUMN 8 HOLDS THE PARAMETERS AFTER THE LASSO STATEMENT.  THESE
    * INCLUDE "AFTER", "BEFORE", "SCREEN", "DEBUG"

    __LASSO&lcPlatform[lnLassoCnt,8]=UPPER(WORDSEARCH(c_LASSO,"m.lcThisLasso"))

*** CHANGE
    * LASSOS FROM A FILE HAVE NO DIMENSIONS (HEIGHT,WIDTH), THEREFORE IT
    * IS NECESSARY TO MAKE THE LASSO A "SCREEN" LASSO.  
    *
    * A LASSO FILE DEFINED IN THE CONFIG.FP WILL BE APPLIED TO ALL SCREENS
    * IN A SCREENSET AUTOMATICALLY, THEREFORE IT CANNOT BE DEFINED AS A 
    * SCREENSET LASSO
    *
    
    IF pcLASSOFILE $ "CONFIG,SETUP,"
       __LASSO&lcPlatform[lnLassoCnt,8]=__LASSO&lcPlatform[lnLassoCnt,8]+" SCREEN "
       IF pcLASSOFILE="CONFIG" .AND. "SCREENSET" $ __LASSO&lcPlatform[lnLassoCnt,8]
          __LASSO&lcPlatform[lnLassoCnt,8]=STRTRAN(__LASSO&lcPlatform[lnLassoCnt,8],"SCREENSET","")
       ENDIF
    ENDIF
*** ENDCHANGE

    * COLUMN 7 HOLDS WHETHER THE ITEMS IN THE LASSO COMMENT SNIPPET SHOULD
    * BE APPLIED TO THE BEGINNING OR TO THE END OF THE WORKING OBJECT
    * COMMENT SNIPPET
   
    __LASSO&lcPlatform[lnLassoCnt,7]=__LASB4&lcPlatform 
    
    IF "BEFORE" $ __LASSO&lcPlatform[I,8]
       __LASSO&lcPlatform[lnLassoCnt,7]=.T.
    ENDIF

    IF "AFTER" $ __LASSO&lcPlatform[I,8]
       __LASSO&lcPlatform[lnLassoCnt,7]=.F.
    ENDIF

    * REMOVE *:LASSO DIRECTIVE 
    * CONVERT *:LASSOADDBRACES ITEMS 
    * REMOVE SPACES, CR, AND LF FROM BEGINNING OF WHAT WILL BECOME "WHAT
    *   GETS APPLIED TO THE WORKING OBJECT COMMENT's SNIPPET" IF IT MEETS
    *   THE LASSO CRITERIA
   
    lcThisLasso=RemLine(c_LASSO,"M.lcThisLasso")
    lcThisLasso=AddBraces(lcThisLasso,c_LASADDBRACES)
    __LASSO&lcPlatform[lnLassoCnt,5]=SUBSTR(lcThisLasso,IGNORE(m.CR_LF+" ",lcThisLasso))   

    * IF *:LASSOSAVE IS IN COMMENT SNIPPET, THEN THE LASSO OBJECT WILL 
    * REMAIN ON THE SCREEN, AND THEREFORE MIGHT BE PROCESSED BY ITSELF
    * THIS MUST BE PREVENTED 
    * STUFF "*-:LASSO" FOLLOWED BY THE ARRAY ROW OF THE LASSO IN 
    * __LASSO INTO THE BOTTOM OF THE COMMENT SNIPPET 
    *    EX.  *-:LASSO1 IS PUT INTO THE COMMENT SNIPPET OF AN OBJECT
    *         JUST PROCESSED BY __LASSO[1] 
    
    IF llSaveLasso
       =WORDSTUFF(M.cr_lf+m.m_LASSO+ALLTRIM(STR(lnLassoCnt))+M.cr_lf,.T.,.F.)
    ENDIF

NEXT I

*** CHANGE: LASSOS FROM A FILE HAVE NO SCREEN COORDINATES AND THEREFORE 
***         CANNOT BE "SAVED" OR "DEBUGGED", SO IGNORE *:LASSOSAVE
***         PART 1 "IF"

IF pcLASSOFILE="NORMAL"

*** ENDCHANGE

   IF __LASDB&lcPlatform .OR. llDebugLasso
      * BLANK OUT VALID, WHEN, RANGELO, AND RANGEHI
      * BLANK OUT COMMENT AND REPLACE WITH "*:NOLASSO" SO THAT ITEM IS IGNORED
      * BY LASSOS, AND "*:NO3D" SO THAT THE LASSOS ARE NOT MADE 3D
      * DISABLE ITEM BY REPLACING DISABLE WITH .T.
      REPLACE COMMENT WITH m.c_NOLASSO+m.cr_lf+"*:NO3D"+m.cr_lf, ;
              VALID WITH "", ;
              WHEN WITH "", ;
              RANGELO WITH "", ;
              RANGEHI WITH "", ;
              DISABLED WITH .T.
   ELSE
      IF !llSaveLasso
         * OTHERWISE:
         * STUFF *:DELETE INTO THE TOP OF THE COMMENT SNIPPET SO THAT LASSO
         * DOES NOT APPEAR ON FINAL SCREEN.
         *
         * ALSO TO MAKE THE DRIVER MORE "FRIENDLY" TO Steven Black's INTL
         * THE RECORD IS DELETED.   
      
         =WORDSTUFF("*:DELETE"+M.cr_lf,.T.,.T.)
         DELETE
      ENDIF   
   ENDIF

*** CHANGE: LASSOS FROM A FILE HAVE NO SCREEN COORDINATES AND THEREFORE 
***         CANNOT BE "SAVED" OR "DEBUGGED"
***         PART 2 "ENDIF"

ENDIF

*** ENDCHANGE

FUNCTION ADDBRACES
PARAMETER pcSNIPPET, pcDirective
* PURPOSE:  this routine looks for the string pcDirective within the string
*           pcSNIPPET.  It then replaces pcDirective with "{{" and then
*           puts "}} at the end of the line.  It does this for all 
*           lines beginning with pcDirectives.
*
* Example if pcDirective="*:BRACE_THIS" and pcSNIPPET contains:
*
*  *:BRACE_THIS DATE()
*  *:BRACE_THIS &. REPLACE disabled with .T.
*
* this function returns:
*
*  {{ DATE() }}
*  {{ &. REPLACE disabled WITH .T }}

PRIVATE llDone,lnCRLFpos

DO WHILE pcDirective $ pcSNIPPET
   llDone=.F.
   lnCRLFpos=AT(pcDirective,pcSNIPPET)+LEN(pcDirective)
   DO WHILE lnCRLFpos<=LEN(pcSNIPPET)
       IF SUBSTR(pcSNIPPET,lnCRLFpos,1) $ M.CR_LF
          pcSNIPPET= SUBSTR(pcSNIPPET,1,lnCRLFpos-1)+"}}"+ ;
                     SUBSTR(pcSNIPPET,lnCRLFpos)
          llDone=.T.
          EXIT       
       ENDIF
       lnCRLFpos=lnCRLFpos+1
   ENDDO  
   IF !llDone
      pcSNIPPET=pcSNIPPET+"}}"
   ENDIF
   pcSNIPPET=STRTRANC(pcSNIPPET,pcDirective,"{{",1,1)
ENDDO
RETURN pcSNIPPET

FUNCTION IGNORE
PARAMETERS pcIgnoreStr,pcSNIPPET

* PURPOSE:  THIS FUNCTION RETURNS THE POSITION OF THE FIRST CHARACTER IN
*           pcSNIPPET THAT IS NOT CONTAINED IN pcIgnoreStr
*                  EX. X=IGNORE(m.CR_LF+" ",COMMENT)
*                  X WILL EQUAL THE POSITION OF THE 
*                  FIRST NON-CR, NON-LF, NON-" " CHARACTER IN THE COMMENT
*                  SNIPPET

PRIVATE lnFirstPos

lnFirstPos=1

DO WHILE SUBSTR(pcSNIPPET,lnFirstPos,1) $ pcIgnoreStr .AND. ;
         lnFirstPos<=LEN(pcSNIPPET)
   lnFirstPos=lnFirstPos+1
ENDDO
RETURN IIF(lnFirstPos<=LEN(pcSNIPPET),lnFirstPos,0)

FUNCTION REMLINE
PARAMETERS pcStartWith, pcSNIPPET, pcRemLine

* PURPOSE: THIS FUNCTION RETURNS pcSNIPPET MINUS ENTIRE LINE BEGINNING WITH
*          pcStartWith.  IT RETURNS pcSNIPPET IF pcStartWith IS NOT IN
*          pcSNIPPET.  THE REMOVED LINE WILL BE STORED TO pcRemLine WHICH 
*          MUST BE PASSED BY REFERENCE
*                  EX. X=REMLINE(c_LASSOINCLUDE,"lcCutOut",@lcRemLine) 

PRIVATE DUMMY,lcParam, lnLineLen, lnLineStart
m.dummy=0
lcParam=WORDSEARCH(pcStartWith,pcSNIPPET)

IF lcParam=CHR(0)
   pcRemLine=""
   RETURN &pcSNIPPET
ENDIF

lnLineLen=LEN(pcStartWith)+LEN(ALLTRIM(lcParam))
lnLineStart=WORDSEARCH(pcStartWith,pcSNIPPET,.F.,@m.dummy,1)

DO WHILE !(SUBSTR(&pcSNIPPET,lnLineStart+lnLineLen,1) $ M.CR_LF) .AND. ;
         (lnLineStart+lnLineLen)<=LEN(&pcSNIPPET)
   lnLineLen=lnLineLen+1      
ENDDO
IF (lnLineStart+lnLineLen)>LEN(&pcSNIPPET)
   lnLineLen=lnLineLen-1
ENDIF

pcRemLine=SUBSTR(&pcSNIPPET,lnLineStart,lnLineLen)

RETURN SUBSTR(&pcSNIPPET,1,lnLineStart-1)+ ;
       SUBSTR(&pcSNIPPET,lnLineStart+lnLineLen)

FUNCTION CnvrtWrd
PARAMETER pcList

DIMENSION laObj[13,2]
laObj[1,1]="TEXT"
laObj[1,2]="5"
laObj[2,1]="BOX"
laObj[2,2]="7"
laObj[3,1]="LIST"
laObj[3,2]="11"
laObj[4,1]="PUSHBUTTON"
laObj[4,2]="12"
laObj[5,1]="RADIOBUTTON"
laObj[5,2]="13"
laObj[6,1]="CHECKBOX"
laObj[6,2]="14"
laObj[7,1]="SAY"
laObj[7,2]="15.0"
laObj[8,1]="GET"
laObj[8,2]="15.1"
laObj[9,1]="EDITREGION"
laObj[9,2]="15.2"
laObj[10,1]="POPUP"
laObj[10,2]="16"
laObj[11,1]="PICTURE"
laObj[11,2]="17"
laObj[12,1]="INVISIBLEBUTTON"
laObj[12,2]="20"
laObj[13,1]="SPINNER"
laObj[13,2]="22"

lcAlphabet="ABCDEFGHIJKLMNOPQRSTUVWXYZ"
lcNumeric="0123456789."

* Filter out numerics from list and remove double commas

lcParseString=UPPER(CHRTRAN(pcList,lcNumeric,""))

DO WHILE ",," $ lcParseString
   lcParseString=STRTRAN(lcParseString,",,",",")
ENDDO

* Remove leading comma if one exists

IF LEFT(lcParseString,1)="," .AND. !EMPTY(STRTRAN(lcParseString,",",""))
   lcParseString=SUBSTR(lcParseString,2)
ENDIF

lcParseString=lcParseString+IIF(RIGHT(lcParseString,1)=",","",",")
lcRetVal=""

DO WHILE !EMPTY(STRTRAN(lcParseString,",",""))
   lcToken=IIF(AT(",",lcParseString)=0, ;
              lcParseString, ;
              SUBSTR(lcParseString,1,AT(",",lcParseString)-1))
   lnTokenLoc=ASCAN(laObj,lcToken,1)           
   IF lnTokenLoc != 0
      lnObjIndex=ASUBSCRIPT(laObj,lnTokenLoc,1)
      lcRetVal=lcRetVal+laObj[lnObjIndex,2]+","
   ELSE
      =WARNING(ALLTRIM(lcToken)+" is not a valid INCLUDE/EXCLUDE object type")
      =CLEANUP()
      CANCEL
   ENDIF
   IF AT(",",lcParseString)=LEN(lcParseString) 
      lcParseString=""
      LOOP
   ELSE
      lcParseString=SUBSTR(lcParseString,AT(",",lcParseString)+1)
   ENDIF                     
ENDDO

lcTemp=UPPER(CHRTRAN(pcList,lcAlphabetic,""))

DO WHILE ",," $ lcTemp
   lcTemp=STRTRAN(lcTemp,",,",",")
ENDDO

IF LEFT(lcTemp,1)="," .AND. !EMPTY(STRTRAN(lcTemp,",",""))
   lcTemp=SUBSTR(lcTemp,2)
ENDIF

IF RIGHT(lcTemp,1)=","
   lcTemp=SUBSTR(lcTemp,1,LEN(lcTemp)-1)
ENDIF

lcRetVal=lcRetVal+lcTemp

RETURN lcRetVal
